import flask

from . import controllers as ctl


auth_blueprint = flask.Blueprint("auth", __name__, url_prefix="/auth")


@auth_blueprint.route("/callback", methods=["GET"])
def callback():
    return ctl.callback.handle_callback()


@auth_blueprint.route("/register", methods=["GET"])
def register():
    return ctl.signup_login.handle_signup()


@auth_blueprint.route("/login", methods=["GET"])
def login():
    return ctl.signup_login.handle_login()


@auth_blueprint.route("/logout", methods=["GET"])
def logout():
    return ctl.logout.handle_logout()


@auth_blueprint.route("/refresh", methods=["POST"])
def refresh():
    return ctl.refresh.handle_refresh()


@auth_blueprint.route("/user", methods=["GET"])
def user():
    return ctl.user.handle_user()



import base64

ENCODING = "ascii"


def encode(string: str) -> str:
    str_bytes = string.encode(ENCODING)
    b64_bytes = base64.b64encode(str_bytes)
    return b64_bytes.decode(ENCODING)


def decode(string: str) -> str:
    b64_bytes = string.encode(ENCODING)
    str_bytes = base64.b64decode(b64_bytes)
    return str_bytes.decode(ENCODING)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A server to handle the OAuth token exchange with FusionAuth.
"""
__version__ = "0.0.1"
__app_name__ = __name__  # for use in __main__ to avoid override



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=invalid-name

import argparse

from . import (
    __name__ as app_name,
    __version__,
    server,
    blueprints,
    config,
)


def cli():
    """
    texserv [-h] [-v] {run} ...

    A server to handle the OAuth token exchange with FusionAuth.

    [options]:
    -h, --help     show this help message and exit
    -v, --version  show program's version number and exit

    {commands}:
        run [-h] (-d | -p)  start the server and listen forever
            -h, --help      show this help message and exit
            -d, --dev       run the app using Flask's dev server
            -p, --prod      run the server for production with Gunicorn
    """
    parser = argparse.ArgumentParser(
        prog=app_name,
        description="App to handle the OAuth token exchange with FusionAuth.",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s v{__version__}",
    )

    subparsers = parser.add_subparsers(
        title="commands",
        help="command help",
    )

    run_cmd = subparsers.add_parser(
        "run",
        description="start the server and run forever",
        help="runs the server",
    )
    run_group = run_cmd.add_mutually_exclusive_group(
        required=True,
    )
    run_group.add_argument(
        "-d",
        "--dev",
        action="store_true",
        help="Run the server in development using Flask's debug mode",
    )
    run_group.add_argument(
        "-p",
        "--prod",
        action="store_true",
        help="Run the server for production with Gunicorn",
    )
    run_cmd.set_defaults(
        func=run,
    )

    args = parser.parse_args()

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


def run(args: argparse.Namespace):
    """
    Start the server and run forever
    """
    if args.prod:
        config.app_config["mode"] = "production"

    app = server.ExchangeServer(
        app_name,
        config.app_config,
        blueprints.app_blueprints,
    )

    app.listen()



import flask
import flask_cors
import gunicorn.app.base


class ExchangeServer(gunicorn.app.base.BaseApplication):
    """
    Texserv is the top-level class managing the application

    Args:
        name:       Name of the application
        config:     Holds values from os.environ
        blueprints: A list of Flask blueprint instances
    """

    def __init__(
        self,
        name: str,
        config: dict[str, str],
        blueprints: list[flask.Blueprint],
    ):
        self._app = flask.Flask(name)
        self._config = config
        self.register_blueprints(blueprints)
        flask_cors.CORS(self._app, supports_credentials=True)
        super().__init__()

    def listen(self):
        if self._config["mode"] == "production":
            self.run()
        else:
            self._app.run(
                debug=True,
                host=self._config["host"],
                port=self._config["port"],
            )

    def load(self):
        return self._app

    def load_config(self):
        gunicorn_config = {
            key.lower(): value
            for key, value in self._config.items()
            if key in self.cfg.settings and value is not None
        }

        for key, value in gunicorn_config.items():
            self.cfg.set(key, value)

    def register_blueprints(self, blueprints: list[flask.Blueprint]):
        if blueprints is not None:
            for blueprint in blueprints:
                self._app.register_blueprint(blueprint)



import urllib.parse

import flask

import authexchange.config as cfg

COOKIE_KEYS = [
    "app.at",
    "app.at_exp",
    "app.rt",
    "app.idt",
]


def handle_logout():
    req = flask.request
    query = urllib.parse.urlencode(
        {
            "post_logout_redirect_uri": req.args.get(
                "post_logout_redirect_uri",
            ),
            "client_id": req.args.get(
                "client_id",
            ),
            "id_token_hint": req.cookies.get(
                "app.idt",
            ),
        }
    )
    redirect_url = "".join(
        [
            cfg.FUSIONAUTH_BASE_URL,
            "/oauth2/logout",
            "?",
            query,
        ]
    )
    res = flask.redirect(redirect_url, code=302)

    for key in COOKIE_KEYS:
        res.delete_cookie(key)

    return res



import urllib.parse

import flask
import pkce

from authexchange.utils import state
import authexchange.config as cfg


def handle_signup():
    return handle_signup_login("register")


def handle_login():
    return handle_signup_login("authorize")


def handle_signup_login(authtype: str):
    """
    Except for the url, the process for signup and login are identical

    Args
        authtype: Either 'register' to signup or 'authorize' to login

    Returns
        A Flask Response that redirects to the corresponding FusionAuth page

    """
    req = flask.request
    new_state = state.push_redirect_url(
        req.args.get("redirect_uri", default=req.host_url),
        req.args.get("state", default=""),
    )
    code_verifier, code_challenge = pkce.generate_pkce_pair()
    redirect_uri = "".join(
        [
            req.scheme,
            "://",
            req.host,
            "/auth/callback",
        ]
    )
    query = urllib.parse.urlencode(
        {
            "response_type": "code",
            "scope": "openid offline_access",
            "client_id": req.args.get("client_id"),
            "state": new_state,
            "redirect_uri": redirect_uri,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
        }
    )
    redirect_url = "".join(
        [
            cfg.FUSIONAUTH_BASE_URL,
            f"/oauth2/{authtype}",
            "?",
            query,
        ]
    )
    res = flask.redirect(
        redirect_url,
        code=302,
    )
    res.set_cookie(
        "code_verifier",
        code_verifier,
        secure=True,
        httponly=True,
        samesite="lax",
    )
    return res



import os

FUSIONAUTH_BASE_URL = os.environ.get("FUSIONAUTH_BASE_URL", "http://localhost:9011")
FUSIONAUTH_NET_URL = os.environ.get("FUSIONAUTH_NET_URL", FUSIONAUTH_BASE_URL)
FUSIONAUTH_CLIENT_ID = os.environ.get("FUSIONAUTH_CLIENT_ID", None)
FUSIONAUTH_CLIENT_SECRET = os.environ.get("FUSIONAUTH_CLIENT_SECRET", None)

mode = os.environ.get("MODE", "development")
host = os.environ.get("HOST", "localhost")
port = os.environ.get("PORT", "9000")
workers = os.environ.get("NUM_WORKERS", "4")

app_config = {
    "mode": mode,
    "host": host,
    "port": port,
    "bind": f"{host}:{port}",
    "workers": workers,
}



import time
import urllib.parse

import flask
import requests

import authexchange.config as cfg


def handle_refresh():
    req = flask.request
    refresh_token = req.cookies.get("app.rt")

    if not refresh_token:
        res = flask.jsonify(
            {"error": "Refresh token is missing"},
        )
        res.status_code = 503
        return res

    fusionauth_res = requests.post(
        url="".join([cfg.FUSIONAUTH_NET_URL, "/oauth2/token"]),
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
        data=urllib.parse.urlencode(
            {
                "grant_type": "refresh_token",
                "refresh_token": req.cookies.get("app.rt"),
                "access_token": req.cookies.get("app.at"),
                "client_id": cfg.FUSIONAUTH_CLIENT_ID,
                "client_secret": cfg.FUSIONAUTH_CLIENT_SECRET,
            }
        ),
        timeout=10,
    )
    data = fusionauth_res.json()

    access_token = data["access_token"]
    refresh_token = data["refresh_token"]
    id_token = data["id_token"]
    expires_in = data["expires_in"]

    if not access_token or not refresh_token:
        res = flask.jsonify(
            {"error": "Either refresh token or access token is missing"},
        )
        res.status_code = 503
        return res

    res = flask.make_response()
    res.status_code = 204

    # set token expiration in a readable cookie
    time_ms = int(time.time() * 1000)
    expires_in_ms = int(expires_in * 1000)
    res.set_cookie(
        "app.at_exp",
        str(time_ms + expires_in_ms),
        httponly=False,
        secure=True,
        samesite="lax",
        max_age=expires_in_ms,
    )
    res.set_cookie(
        "app.at",
        access_token,
        httponly=True,
        secure=True,
        samesite="lax",
    )
    res.set_cookie(
        "app.rt",
        refresh_token,
        httponly=True,
        secure=True,
        samesite="lax",
    )
    res.set_cookie(
        "app.idt",
        id_token,
        httponly=False,
        secure=True,
        samesite="lax",
    )

    # clean up cookie set in login/register
    res.delete_cookie("code_verifier")

    return res



import flask
import requests

import authexchange.config as cfg


def handle_user():
    req = flask.request
    access_token = req.cookies.get("app.at")

    if not access_token:
        res = flask.jsonify({"error": "Access token is missing"})
        res.status_code = 401
        return res

    url = "".join([cfg.FUSIONAUTH_NET_URL, "/oauth2/userinfo"])
    res = requests.get(
        url,
        headers={"Authorization": f"Bearer {access_token}"},
        timeout=10,
    )
    data = res.json()

    return flask.jsonify(data)



from authexchange.utils import b64


def generate_redirect_url(state: str) -> str:
    encoded_uri, *_ = state.split(":")
    redirect_uri = b64.decode(encoded_uri)
    return redirect_uri


def push_redirect_url(redirect_uri: str, state: str) -> str:
    encoded_uri = b64.encode(redirect_uri)
    return ":".join([encoded_uri, state])



import flask

ping_blueprint = flask.Blueprint("ping", __name__, url_prefix="")


@ping_blueprint.route("/ping", methods=["GET"])
def ping():
    """
    healthcheck
    """
    return flask.jsonify({"msg": "pong"})
